from typing import Tuple, Optional

Point = Tuple[int,int]

#Parametri condivisi per la curva y^2=x^3++7x+3
p = 13
n = 13
G = [3,5]
a = 7
b = 3

#Queste due funzioni restituiscono la coordinata x e y di un punto di una curva (rispettivamente)
def x(P : Point) -> int:
    return P[0]

def y(P):
    return P[1]

#Questa funzione somma due punti di una curva ellittica, quindi restituisce P3 = P1 + P2 = (x3, y3)
#in cui il punto può essere indicato dalle coordinate x e y intere oppure da none se è il punto all'infinito
def point_add(P1 : Optional[Point], P2 : Optional[Point]) -> Optional[Point]:
    if P1 is None:
        return P2
    if P2 is None:
        return P1
    if (x(P1) == x(P2)) and (y(P1) != y(P2)):
        return None
    if P1 == P2:
        lam = ((3 * x(P1)**2 + a) * pow(2 * y(P1), p - 2, p)) % p
    else:
        lam = ((y(P2) - y(P1)) * pow(x(P2) - x(P1), p - 2, p)) % p

    x3 = (lam**2 - x(P1) - x(P2)) % p
    y3 = (lam * (x(P1) - x3) - y(P1)) % p

    return x3, y3

#Questa funzione calcola il punto R = dP 
def point_mul(d : int, P : Point) -> Point:
    R = None

    for i in range(256):
        if (d >> i) & 1:
            R = point_add(R,P)

        P = point_add(P,P)

    return R

#Questa funzione restituisce una chiave pubblica a partire da una chiave privata
def pubkey_gen(private_key : int) -> Point:
    R = point_mul(private_key, G)

    assert R is not None
    return R

#Questa funzione serve per controllare che il punto trovato 
#appartenga alla curva ellittica:    
def isPoint(P):
    sx=(y(P)**2) % p
    dx=((x(P)**3)+a*x(P)+b) % p
    if(sx==dx):
        return True
    else:
        return False    

#Adesso possiamo iniziare il protocollo di scambio di chiavi di Diffie-Hellman
print("Inserire la chiave privata di Alice")
private_key_a=int(input())
print("Inserire la chiave privata di Bob")
private_key_b=int(input())

pubblic_key_a=pubkey_gen(private_key_a)
print("La chiave pubblica di Alice è: %s" %(pubblic_key_a,))
print("Controllo che appartenga alla curva ellittica:")
print(isPoint(pubblic_key_a)) 

pubblic_key_b=pubkey_gen(private_key_b)     
print("La chiave pubblica di Bob è: %s" %(pubblic_key_b,))
print("Controllo che appartenga alla curva ellittica:")
print(isPoint(pubblic_key_b))  

shared_key_a=point_mul(private_key_a,pubblic_key_b)
shared_key_b=point_mul(private_key_b,pubblic_key_a)
print("La chiave condivisa calcolata da Alice è: %s" %(shared_key_a,))
print("La chiave condivisa calcolata da Bob è: %s" %(shared_key_b,))

print("Controllo che appartenga alla curva ellittica:")
print(isPoint(shared_key_a))  



