#Prima di tutto definiamo il modulo di numeri comlessi:
from typing import Tuple, Optional
import networkx as nx  
import matplotlib.pyplot as plt
from numpy import random

Point = Tuple[complex,complex]

def mod(x:complex,p:int):
    rep=(x.real)%p
    imp=(x.imag)%p
    return (complex(rep,imp))
    
#adesso vogliamo definire la moltiplicazione modulo p tra numeri complessi
    
def mol(x1:complex,x2:complex,p:int):
    return(mod((x1*x2),p))

#adesso si vuole definire la potenza, per farlo 
#occorre utilizzare un algoritmo adatto:

def pot(x:complex,d:int,p:int):
    R=1
    for i in range(256):
        if (d >> i) &1:
            R = mol(x,R,p)
        x = mol(x,x,p)
    return (R)    

#Serve anche definire il contrario di un numero complesso
# a+bi cioè a-bi    
  
def contr(x:complex,p):
    re=x.real
    im=-x.imag
    return mod((complex(re,im)),p)

#adesso per semplicità definiamo l'inversa
    
def inv(x,p):
    contrario=contr(x,p)
    potenza=pot(((x.real)**2+(x.imag)**2),p-2,p)
    return (mol(contrario,potenza,p))

#implementazione della duplicazione di un punto
    
def dup(x,p,a):
    x2=mol(x,x,p)
    x21=mod((x2-1),p)
    num=mol(x21,x21,p)
    ax=mol(x,a,p)
    xper4=mol(x,4,p)
    den1=mod((x2+ax),p)
    den2=mod((den1+1),p)
    den3=mol(den2,xper4,p)
    den4=inv(den3,p)
    return(mol(num,den4,p))

#Queste due funzioni restituiscono la coordinata x e y di un punto di una curva (rispettivamente)
def x(P : Point) -> int:
    return P[0]

def y(P):
    return P[1]

#Questa funzione somma due punti di una curva ellittica, cioé ci restituisce P3 = P1 + P2 = (x3, y3)
#Optional[x] significa che l'elemento x  a cui ci si sta riferendo è di tipo Point OPPURE None
    
def point_add(P1 : Optional[Point], P2 : Optional[Point],p) -> Optional[Point]:
    if P1 is None:
        return P2
    if P2 is None:
        return P1
    if (x(P1) == x(P2)) and (y(P1) != y(P2)):
        return None
    if P1 == P2:
        lam = mod(((3 * x(P1)**2 + 2*a*x(P1)+1) * inv(2 * y(P1),p)),p)
    else:
        lam =mod( ((y(P2) - y(P1)) * inv(x(P2) - x(P1), p)),p)

    x3 = mod((lam**2 - x(P1) - x(P2)-a),p)
    y3 = mod((lam * (x(P1) - x3) - y(P1)),p)

    return x3, y3

#Questa funzione calcola il punto R = dP
def point_mul(d : int, P : Point,p) -> Point:
    R = None

    for i in range(256):
        if (d >> i) & 1:
            R = point_add(R,P,p)

        P = point_add(P,P,p)

    return R

#Questa funzione restituisce una chiave pubblica a partire da una chiave privata
def pubkey_gen(private_key : int,G) -> Point:
    R = point_mul(private_key, G)
    assert R is not None
    return R

#questa funzione permette di assegnare una variabile di tipo Point

def ass(x:complex,y:complex)->Point:
    return x,y

# questa funzione permette di ottenere il nuovo a di una curva data un isogenia 
# di grado 2    

def new_a(x:complex,p)->complex:
    a_prime=mol(mod(1-mol(pot(x,2,p),2,p),p),2,p)
    return a_prime

# questa funzione permette di ottenere il nuovo a di una curva data un isogenia 
# di grado 3    

def new_aB(x:complex,p,a)->complex:
    ax=mol(a,x,p)
    x2=pot(x,2,p)
    x26=mol(6,x2,p)
    a_prime=mol(mod(ax-x26+6,p),x,p)
    return a_prime

# Questa funzione restituisce il j-invariante di una curva

def j_inv(a:complex,p)->complex:
    A2=pot(a,2,p)
    A32=mod(A2-3,p)
    A323=pot(A32,3,p)
    num=mol(256,A323,p)
    den=inv(mod(A2-4,p),p)
    return mol(num,den,p)
      
# Questa funzione calocla la radice quadrata di un numero complesso modulo p

def root(x,p,q):
    a1=pot(x,107,p)
    alpha=mol(a1,mol(a1,x,p),p)
    a0=pot(alpha,q+1,p)
    if a0==-1:
        return None
    x0=mol(a1,x,p)
    if alpha==-1:
        x=complex(0,x0)
    else:
        b=pot(mod(1+alpha,p),215,p)
        x=mol(b,x0,p)
    return x

# Questa funzione calcola il parametro c dell'isogenia di grado 2

def calcolo_c_A(P:Point,p,a,xr):
    x=P[0]
    l=mol(x,xr,p)
    l=mod(l-1,p)
    l=mol(l,x,p)
    den=inv(mod(x-xr,p),p)
    x_new=mol(l,den,p)
    y=P[1]
    x3_new=pot(x_new,3,p)
    x2_new=pot(x_new,2,p)
    ax2_new=mol(a,x2_new,p)
    arg=mod((x3_new+ax2_new+x_new),p)
    rad1=root(arg,p,p**2)
    rad2=mol(-1,rad1,p)  
    print(rad1,rad2)
    x2=pot(x,2,p)
    xrx=mol(xr,x,p)
    xrx2=mol(xrx,2,p)
    fprimo=mol(mol(xr,x2-xrx2+1,p),inv(mol(x-xr,x-xr,p),p),p)
    c1=mol(rad1,inv(mol(y,fprimo,p),p),p)
    c2=mol(rad2,inv(mol(y,fprimo,p),p),p) 
    minimum=min(abs(c1),abs(c2))
    if minimum==abs(c1):
        return c1
    else:
        return c2

# Questa funzione permette di calcolare dove viene mandato un punto 
# tramite un isogenia di grado 2        
    
def PhiA_i(P:Point,p,a,xr,c):   
    x=P[0]
    l=mol(x,xr,p)
    l=mod(l-1,p)
    l=mol(l,x,p)
    den=inv(mod(x-xr,p),p)
    x_new=mol(l,den,p)
    y=P[1]
    x2=pot(x,2,p)
    xrx=mol(xr,x,p)
    xrx2=mol(xrx,2,p)
    fprimo=mol(mol(xr,x2-xrx2+1,p),inv(mol(x-xr,x-xr,p),p),p)
    y_prime=mol(c,mol(y,fprimo,p),p)
    P_prime=ass(x_new,y_prime)
    print('le vecchie coordinate erano', P, 'invece le nuove sono', P_prime)
    return P_prime       

def Phix(x,p,a,xr):
    l=mol(x,xr,p)
    l=mod(l-1,p)
    l=mol(l,x,p)
    den=inv(mod(x-xr,p),p)
    x_new=mol(l,den,p)
    return x_new

# Questa funzione calcola il parametro c dell'isogenia di grado 2

def calcolo_c_B(P:Point,p,a,xr):
    x=P[0]
    l=mol(x,xr,p)
    l=mod(l-1,p)
    l=mol(l,l,p)
    l=mol(l,x,p)
    den=inv(pot(mod(x-xr,p),2,p),p)
    x_new=mol(l,den,p)
    y=P[1]
    x3_new=pot(x_new,3,p)
    x2_new=pot(x_new,2,p)
    ax2_new=mol(a,x2_new,p)
    arg=mod((x3_new+ax2_new+x_new),p)
    rad1=root(arg,p,p**2)
    rad2=mol(-1,rad1,p)  
    x2=pot(x,2,p)
    xrx2=mol(xr,x2,p)
    xrx=mol(xr,x,p)
    xr2=pot(xr,2,p)
    xr23=mol(xr2,3,p)
    num=mol((xrx-1),(xrx2+mol(1-xr23,x,p)+xr),p)
    den=inv(pot(x-xr,3,p),p)
    fprimo=mol(num,den,p)
    c1=mol(rad1,inv(mol(y,fprimo,p),p),p)
    c2=mol(rad2,inv(mol(y,fprimo,p),p),p) 
    minimum=min(abs(c1),abs(c2))
    if minimum==abs(c1):
        return c1
    else:
        return c2
 
# Questa funzione permette di calcolare dove viene mandato un punto 
# tramite un isogenia di grado 2      
    
def PhiB_i(P:Point,p,a,xr,c):
    x=P[0]
    l=mol(x,xr,p)
    l=mod(l-1,p)
    l=mol(l,l,p)
    l=mol(l,x,p)
    den=inv(pot(mod(x-xr,p),2,p),p)
    x_new=mol(l,den,p)
    y=P[1]
    x2=pot(x,2,p)
    xrx2=mol(xr,x2,p)
    xrx=mol(xr,x,p)
    xr2=pot(xr,2,p)
    xr23=mol(xr2,3,p)
    num=mol((xrx-1),(xrx2+mol(1-xr23,x,p)+xr),p)
    den=inv(pot(x-xr,3,p),p)
    fprimo=mol(num,den,p)
    y=mol(c,mol(y,fprimo,p),p)
    P_prime=ass(x_new,y)
    print('le vecchie coordinate erano', P, 'invece le nuove sono', P_prime)
    return P_prime   

# Inizia il vero e proprio algoritmo
 
#parametri pubblici iniziali
p=431    
a=423+329j
xPA=248+100j
yPA=199+304j          
xQA=394+426j
yQA=79+51j
xPB=358j+275
yPB=410j+104
xQB=20j+185
yQB=281j+239

PA=ass(xPA,yPA)
QA=ass(394+426j,79+51j)
PB=ass(xPB,yPB)
QB=ass(xQB,yQB)


kA=random.randint(1,15)
kB=random.randint(1,26)

expA=4
expB=3

if kA%2==0:
   kA=kA+1
if kA==3:
   kA=5
   
esp=expA
SA=point_add(PA,point_mul(kA,QA,p),p)
print(SA)
SB=point_add(PB,point_mul(kB,QB,p),p)
j=j_inv(a,p)

print('RAPPRESENTAZIONE DEL GRAFO')
GA=nx.DiGraph()
GA.add_node(j)
nx.draw(GA, with_labels=True, font_weight='5',node_size=3000,font_size=8)
plt.show()


for i in range(1,esp+1,1):

    print('Aggiornamento di A',i)
    print(SA)
    print(p)
    print(2**(esp-i))
    R=point_mul(2**(esp-i),SA,p)
    print('R =', R)
    a=new_a(x(R),p)
    print(a)
    j_new=j_inv(a,p)
    print('RAPPRESENTAZIONE DEL GRAFO')
    GA.add_node(j)
    GA.add_edge(j, j_new)
    nx.draw(GA, with_labels=True, font_weight='5',node_size=3000)
    plt.show()
    c=calcolo_c_A(PB,p,a,x(R))
    print('Aggiornamento di PB')
    PB=(PhiA_i(PB,p,a,x(R),c))
    print('Aggiornamento di QB')
    QB=(PhiA_i(QB,p,a,x(R),c))
    print('Aggiornamento di SA')
    SA=(PhiA_i(SA,p,a,x(R),c))
    j=j_new
    print('il j invariante è dato da',j)

PkA=[a,PB,QB]
print(PkA)

print("-----------------------------------------------------------------------")

esp=expB
a=329j+423
j=190+87j
GB=nx.DiGraph()
GB.add_node(j)
nx.draw(GB, with_labels=True, font_weight='5',node_size=3000,font_size=8)
plt.show()

for i in range(1,esp+1,1):

    print('Aggiornamento di Bob',i)
    R=point_mul(3**(esp-i),SB,p)
    a=new_aB(x(R),p,a)
    j_new=j_inv(a,p)
    print('RAPPRESENTAZIONE DEL GRAFO')
    GB.add_node(j)
    GB.add_edge(j, j_new)
    nx.draw(GB, with_labels=True, font_weight='5',node_size=3000)
    plt.show()
    c=calcolo_c_B(PA,p,a,x(R))
    print('Aggiornamento di PA')
    PA=(PhiB_i(PA,p,a,x(R),c))
    print('Aggiornamento di QA')
    QA=(PhiB_i(QA,p,a,x(R),c))
    print('Aggiornamento di SB')
    SB=(PhiB_i(SB,p,a,x(R),c))
    j=j_new
    print('il j invariante è dato da',j)

PkB=[a,PA,QA]


SA=point_add(PkB[1],point_mul(kA,PkB[2],p),p)
#SB=point_add(PkA[1],point_mul(kB,PkA[2],p),p)
#SA=ass(125j+357,415j+249)

esp=expA
a=PkB[0]
j=j_inv(a,p)
print('RAPPRESENTAZIONE DEL GRAFO')
GA1=nx.DiGraph()
GA1.add_node(j)
nx.draw(GA1, with_labels=True, font_weight='5',node_size=3000,font_size=8)
plt.show()

print("-----------------------------------------------------------------------")

for i in range(1,esp+1,1):

    print('Aggiornamento Alice',i)
    R=point_mul(2**(esp-i),SA,p)
    a=new_a(x(R),p)
    j_new=j_inv(a,p)
    print('RAPPRESENTAZIONE DEL GRAFO')
    GA1.add_node(j)
    GA1.add_edge(j, j_new)
    nx.draw(GA1, with_labels=True, font_weight='5',node_size=3000,)
    plt.show()
    print('il j invariante di A al passo', i, 'è',j)
    c=calcolo_c_A(PkB[1],p,a,x(R))
    PkB[1]=(PhiA_i(PkB[1],p,a,x(R),c))
    PkB[2]=(PhiA_i(PkB[2],p,a,x(R),c))
    SA=(PhiA_i(SA,p,a,x(R),c))
    j=j_new
    print('il j invariante è dato da',j)    
    
j_A=j   
print('il j invariante è dato da',j)    

esp=expB
a=PkA[0]
#PkA=[423j+179, ass(142j+183,119j+360), ass(220j+314, 289j+10)]
SB=point_add(PkA[1],point_mul(kB,PkA[2],p),p)
j=j_inv(a,p)
GB1=nx.DiGraph()
GB1.add_node(j)
nx.draw(GB1, with_labels=True, font_weight='5',node_size=3000)
plt.show()

#SB=ass(393j+124,187j+380)

print("-----------------------------------------------------------------------")

for i in range(1,esp+1,1):

    print('Aggiornamento di Bob',i)
    R=point_mul(3**(esp-i),SB,p)
    a=new_aB(x(R),p,a)
    print('a di Bob al passo', i, 'è',a)
    j_new=j_inv(a,p)
    print('RAPPRESENTAZIONE DEL GRAFO')
    GB1.add_node(j_new)
    GB1.add_edge(j, j_new)
    nx.draw(GB1, with_labels=True, font_weight='3',node_size=3000,font_size=8)
    plt.show()
    print('il j invariante di B al passo', i, 'è',j)
    c=calcolo_c_B(PkA[1],p,a,x(R))
    PkA[1]=(PhiB_i(PkA[1],p,a,x(R),c))
    PkA[2]=(PhiB_i(PkA[2],p,a,x(R),c))
    SB=(PhiB_i(SB,p,a,x(R),c))
    j=j_new
    print('il j invariante è dato da',j)    

print('I due j-invarianti che costituiscono la chiave pubblica condivisa sono:', j_A,j)


